//eit.c
//
//

#ifdef _WIN32
#include <windows.h>
#include <stdio.h>
#include <io.h>
#include <wininet.h>

typedef ULONGLONG ulong64;
typedef ULONG     dword;
typedef USHORT    word;
typedef UCHAR     byte;
#define TAP_Print printf
#else
#include <tap.h>
#endif

#include "exTap.h"
#include "eit.h"

//----------------------------------------------------------------------------
//void dump(byte *buf, int len)
//{
//    int i,j;
//
//    for (i=0; i<len; i+=0x10) {
//        printf("\n%04x : ",i);
//        for (j=0; j<0x10; j++) {
//            if ((i+j)<len) printf("%02x ",buf[i+j]); else printf("   ");
//        }
//        printf("  ");
//        for (j=0; j<0x10; j++) {
//            if ((i+j)<len) printf("%c",to_printable(buf[i+j])); else printf(" ");
//        }
//    }
//}

#define NO_DESCRIPTOR_TAGS   68
EnumType descriptor_str[] = {
    { NO_DESCRIPTOR_TAGS,                       "UNKNOWN"                       },
    { SHORT_EVENT_DESCRIPTOR,                   "SHORT EVENT"                   },
    { EXTENDED_EVENT_DESCRIPTOR,                "EXTENDED EVENT"                },
    { COMPONENT_DESCRIPTOR,                     "COMPONENT"                     },
    { CONTENT_DESCRIPTOR,                       "CONTENT"                       },
    { PARENTAL_RATING_DESCRIPTOR,               "PARENTAL RATING"               },
};

//stream_content, component_type, description
#define NO_COMPONENT_TYPES      34
struct {
    byte   stream_content;
    byte   component_type;
    byte   description[0x80];
} component[NO_COMPONENT_TYPES] = {
    { 0x1,0x01,"4:3" },
    { 0x1,0x02,"16:9, with pan vectors" },
    { 0x1,0x03,"16:9, without pan vectors" },
    { 0x1,0x04,">16:9" },
    { 0x1,0x05,"4:3, 30Hz" },
    { 0x1,0x06,"16:9 with pan vectors, 30Hz" },
    { 0x1,0x07,"16:9 without pan vectors, 30Hz" },
    { 0x1,0x08,">16:9, 30Hz" },
    { 0x1,0x09,"hd, 4:3" },
    { 0x1,0x0a,"hd, 16:9 with pan vectors" },
    { 0x1,0x0b,"hd, 16:9 without pan vectors" },
    { 0x1,0x0c,"hd, >16:9" },
    { 0x1,0x0d,"hd, 4:3, 30Hz" },
    { 0x1,0x0e,"hd, 16:9, with pan vectors, 30Hz" },
    { 0x1,0x0f,"hd, 16:9, without pan vectors, 30Hz" },
    { 0x1,0x10,"hd, >16:9, 30Hz" },

    { 0x2,0x01,"single mono channel" },
    { 0x2,0x02,"dual mono channel" },
    { 0x2,0x03,"stereo (2 channel)" },
    { 0x2,0x04,"multi-lingual, multi-channel" },
    { 0x2,0x05,"surround sound" },
    { 0x2,0x40,"audio description for the visually impaired" },
    { 0x2,0x41,"audio for the hard of hearing" },

    { 0x3,0x01,"EBU Teletext subtitles" },
    { 0x3,0x02,"associated EBU Teletext" },
    { 0x3,0x03,"VBI data" },
    { 0x3,0x10,"DVB subtitles (normal)" },
    { 0x3,0x11,"DVB subtitles (normal) (4:3)" },
    { 0x3,0x12,"DVB subtitles (normal) (16:9)" },
    { 0x3,0x13,"DVB subtitles (normal) (2.21:1)" },
    { 0x3,0x20,"DVB subtitles (for the hard of hearing)" },
    { 0x3,0x21,"DVB subtitles (for the hard of hearing) (4:3)" },
    { 0x3,0x22,"DVB subtitles (for the hard of hearing) (16:9)" },
    { 0x3,0x23,"DVB subtitles (for the hard of hearing) (2.21:1)" }
};

//content_nibble1, content_nibble2
// level 2 does not appear to be used
#define NO_CONTENT_TYPES      80
struct {
    byte nibble1;
    byte nibble2;
    byte description[0x40];
} content[] = {
    { 0x0,0x00,"undefined" },

    { 0x1,0x00,"movie/drama" },
    { 0x1,0x01,"detective/thriller" },
    { 0x1,0x02,"adventure/western/war" },
    { 0x1,0x03,"science fiction/fantasy/horror" },
    { 0x1,0x04,"comedy" },
    { 0x1,0x05,"soap/melodrama/folkloric" },
    { 0x1,0x06,"romance" },
    { 0x1,0x07,"serious/classical/religious/historical movie/drama" },
    { 0x1,0x08,"adult movie/drama" },

    { 0x2,0x00,"news/current affairs" },
    { 0x2,0x01,"news/weather report" },
    { 0x2,0x02,"news magazine" },
    { 0x2,0x03,"documentary" },
    { 0x2,0x04,"discussion/interview/debate" },

    { 0x3,0x00,"show/game show" },
    { 0x3,0x01,"game show/quiz/contest" },
    { 0x3,0x02,"variety show" },
    { 0x3,0x03,"talk show" },

    { 0x4,0x00,"sports" },
    { 0x4,0x01,"special sports events" },
    { 0x4,0x02,"sports magazines" },
    { 0x4,0x03,"football/soccer" },
    { 0x4,0x04,"tennis/squash" },
    { 0x4,0x05,"team sports" },
    { 0x4,0x06,"athletics" },
    { 0x4,0x07,"motor sport" },
    { 0x4,0x08,"water sport" },
    { 0x4,0x09,"winter sports" },
    { 0x4,0x0a,"equestrian" },
    { 0x4,0x0b,"martial sports" },

    { 0x5,0x00,"children's/youth programmes" },
    { 0x5,0x01,"pre-school children's programmes" },
    { 0x5,0x02,"entertainment programmes for 6 to 14" },
    { 0x5,0x03,"entertainment programmes for 10 to 16" },
    { 0x5,0x04,"informational/educational/school programmes" },
    { 0x5,0x05,"cartoons/puppets" },

    { 0x6,0x00,"music/ballet/dance" },
    { 0x6,0x01,"rock/pop" },
    { 0x6,0x02,"serious music/classical music" },
    { 0x6,0x03,"folk/traditional music" },
    { 0x6,0x04,"jazz" },
    { 0x6,0x05,"musical/opera" },
    { 0x6,0x06,"ballet" },

    { 0x7,0x00,"arts/culture (without music)" },
    { 0x7,0x01,"performing arts" },
    { 0x7,0x02,"fine arts" },
    { 0x7,0x03,"religion" },
    { 0x7,0x04,"popular culture/traditional arts" },
    { 0x7,0x05,"literature" },
    { 0x7,0x06,"film/cinema" },
    { 0x7,0x07,"experimental film/video" },
    { 0x7,0x08,"broadcasting/press" },
    { 0x7,0x09,"new media" },
    { 0x7,0x0a,"arts/culture magazines" },
    { 0x7,0x0b,"fashion" },

    { 0x8,0x00,"social/political issues/economics" },
    { 0x8,0x01,"magazines/reports/documentary" },
    { 0x8,0x02,"economics/social advisory" },
    { 0x8,0x03,"remarkable people" },

    { 0x9,0x00,"education/science/factual topics" },
    { 0x9,0x01,"nature/animals/environment" },
    { 0x9,0x02,"technology/natural sciences" },
    { 0x9,0x03,"medicine/physiology/psychology" },
    { 0x9,0x04,"foreign countries/expeditions" },
    { 0x9,0x05,"social/spiritual sciences" },
    { 0x9,0x06,"further education" },
    { 0x9,0x07,"languages" },

    { 0xa,0x00,"leisure hobbies" },
    { 0xa,0x01,"tourism/travel" },
    { 0xa,0x02,"handicraft" },
    { 0xa,0x03,"motoring" },
    { 0xa,0x04,"fitness & health" },
    { 0xa,0x05,"cooking" },
    { 0xa,0x06,"advertisement/shopping" },
    { 0xa,0x07,"gardening" },

    { 0xb,0x00,"original language" },
    { 0xb,0x01,"black and white" },
    { 0xb,0x02,"unpublished" },
    { 0xb,0x03,"live broadcast" },
};

char *dvb_enumtype(EnumType *qenum, int id)
{
    int i;

    for(i=1; i<=qenum[0].id; i++) {
        if (qenum[i].id == id) {
            return qenum[i].description;
        }
    }
    return qenum[0].description;
}

char *get_component_type_str(int scontent, int ctype)
{
    int i;

    for(i=0; i<NO_COMPONENT_TYPES; i++) {
        if ((component[i].stream_content == scontent) && (component[i].component_type == ctype)) {
            return component[i].description;
        }
    }
    return component[0].description;
}

char *get_content_type_str(int nibble1, int nibble2)
{
    int i;

    for(i=1; i<NO_CONTENT_TYPES; i++) {
        if ((content[i].nibble1 == nibble1) && (content[i].nibble2 == nibble2)) {
            return content[i].description;
        }
    }
    return content[0].description;
}
//----------------------------------------------------------------------------
//date stuff
char *day_str[] = {"Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"};

char *getdow(int mjd)
{
    return day_str[((mjd + 2) % 7) + 1];
}

int getmjd(word mjd, byte *day, byte *month, word *year)
{
    int yy,mm,k;

#ifdef _WIN32
    yy = (int)((mjd - 15078.2) / 365.25);
    mm = (int) ((mjd - 14956.1 - (int) (yy * 365.25)) / 30.6001);
    k = ((mm == 14) || (mm == 15)) ? 1 : 0;
    *day = (byte)(mjd - 14956 - (int)(yy * 365.25) - (int)(mm * 30.6001));
    *month = (byte)(mm - 1 - k * 12);
    *year = (word)(1900 + yy + k);
#else
    byte weekDay;

    TAP_ExtractMjd(mjd, year, month, day, &weekDay);
#endif
    return 0;
}

char mjdstr[0x20];

char *mjd_to_string(int mjd)
{
    byte day, month;
    word year;

    getmjd(mjd, &day, &month, &year);
    sprintf(mjdstr, "%02d/%02d/%04d", day, month, year);
    return mjdstr;
}

int setmjd(int day, int month, int year)
{
    int k;
    int mjd;

#ifdef _WIN32
    k = ((month == 1) || (month == 2)) ? 1 : 0;
    mjd = 14956 + day + (int)((year - 1900 - k) * 365.25) + (int)((month + 1 + k * 12) * 30.6001);
#else
    mjd = (int)TAP_MakeMjd((word)year, (byte)month, (byte)day);
#endif
    return mjd;
}

//----------------------------------------------------------------------------
//crc stuff
//*
//* The FCS-32 generator polynomial: x**0 + x**1 + x**2 + x**4 + x**5
//*                      + x**7 + x**8 + x**10 + x**11 + x**12 + x**16
//*                      + x**22 + x**23 + x**26 + x**32.
//*
#define CRC_POLYNOMIAL  0xedb88320

//*
//* FCS lookup table as calculated by the table generator.
//*
static dword fcstab8[0x100];

static dword bitrevN(int n, dword data)
{
    int i;
    dword result;

    result = 0;
    for (i=0; i<n; i++) {
        result += ((data >> (n-1-i)) & 1) << (i);
    }
    return result;
}

void init_crc(void)
{
    int b;
    int b2;
    dword v;
    int i;

    for (b=0; b<(1<<8); b++) {
        b2 = bitrevN(8,b);
        v = (dword)b;
        for(i=0; i<8; i++) {
            v = (v >> 1) ^ ((v & 1) ? CRC_POLYNOMIAL : 0);
        }

        fcstab8[b2] = bitrevN(32,v);
    }
}

// 32 bit crc, 8 bit data, 8 bits at a time
// lookup table = 2^8=256 words
dword crc8(dword crc, byte *data, int len)
{
    int i;
    int index;
    dword dataout32;

    for(i=0; i<len; i++) {
        index = ((crc>>24) ^ (data[i])) & 0xff;
        dataout32 = fcstab8[index];
        crc = (crc << 8) ^ dataout32;
    }
    return crc;
}

dword crc32(byte *buf, int blen)
{
    dword crc;

    crc = 0xffffffff;
    crc = crc8(crc, buf, blen);
    return crc;
}

//----------------------------------------------------------------------------
int set_short_event_descriptor(type_eventtable *et, byte *pkt)
{
    int i,j;
    byte descriptor_tag = SHORT_EVENT_DESCRIPTOR;
    byte descriptor_length;
    byte iso_639_language_code[3] = {'e','n','g'};
    int text_length;
    int len;

    //descriptor length must be < 0x100
    //just chop the text ...

    //name length is in et->event_name_length
    //description is zero terminated
    if (et->event_name_length > 0xfc) {
        et->event_name_length = 0xfc;
    }
    text_length = strlen(&et->event_name[et->event_name_length]);
    len = 5 + et->event_name_length + text_length;
    if (len > 0x101) {
        et->event_name[0x101] = '\0';
    }

    j = 0;
    pkt[j++] = descriptor_tag;
    j++;  //fill in descriptor_length later
    pkt[j++] = iso_639_language_code[0];
    pkt[j++] = iso_639_language_code[1];
    pkt[j++] = iso_639_language_code[2];

    pkt[j++] = et->event_name_length;
    memcpy(&pkt[j], et->event_name, et->event_name_length);
    j += et->event_name_length;
    pkt[j++] = (byte)text_length;
    memcpy(&pkt[j], &et->event_name[et->event_name_length], text_length);
    j += text_length;

    if (j > 0x101) {
        TAP_Print("\nERROR: oops; short descriptor too long ...");
        dump((dword)pkt, j);
    }
	descriptor_length = j-2;
	pkt[1] = descriptor_length;
    return j;
}

int set_extended_event_descriptor(type_eventtable *et, byte *pkt)
{
    int i,j,l;
    byte descriptor_tag = EXTENDED_EVENT_DESCRIPTOR;
    byte descriptor_length;
    byte descriptor_number = 0;
    byte last_descriptor_number = 0;
    byte iso_639_language_code[3] = {'e','n','g'};
    int length_of_items_pos;
    byte length_of_items;
    int len;
    int offset;

    j = 0;
    pkt[j++] = descriptor_tag;
    j++;  //fill in descriptor_length later
    pkt[j++] = ((descriptor_number << 4) & 0xf0)
             | ((last_descriptor_number << 0) & 0x0f);
    pkt[j++] = iso_639_language_code[0];
    pkt[j++] = iso_639_language_code[1];
    pkt[j++] = iso_639_language_code[2];
    length_of_items_pos = j++;  //fill in length_of_items later

    //descriptor length must be < 0x100
    //just chop the text ...
    if (et->extended_length > 0xfc) {
        et->extended_length = 0xfc;
        et->extended_event_name[et->extended_length] = '\0';
        et->extended_event_name[et->extended_length+1] = '\0';
    }

    //zero terminated pairs
    l = j;
    offset = 0;
    for (i=0; i<et->num_extended; i++) {
        len = strlen(&et->extended_event_name[offset]);
        pkt[j++] = len;
        memcpy(&pkt[j], &et->extended_event_name[offset], len);
        j += len;
        offset ++;                      //skip the zero terminator
        offset += len;

        len = strlen(&et->extended_event_name[offset]);
        pkt[j++] = len;
        memcpy(&pkt[j], &et->extended_event_name[offset], len);
       j += len;
        offset ++;                      //skip the zero terminator
        offset += len;
    }
    //
    length_of_items = j-l;
    pkt[length_of_items_pos] = length_of_items;

    //there may be some plain text
    if (et->extended_event_name) {
        len = strlen(&et->extended_event_name[offset]);
        pkt[j++] = len;
        memcpy(&pkt[j], &et->extended_event_name[offset], len);
        j ++;
        j += len;
        offset ++;
        offset += len;
    }

	descriptor_length  = j-2;
	pkt[1] = descriptor_length;
    return j;
}

int set_parental_rating_descriptor(type_eventtable *et, byte *pkt)
{
    int j;
    byte descriptor_tag = PARENTAL_RATING_DESCRIPTOR;
    byte descriptor_length;
    byte country_code[3] = {'e','n','g'};

	j = 0;
	pkt[j++] = descriptor_tag;
	j++;  //fill in descriptor_length later
	pkt[j++] = country_code[0];
	pkt[j++] = country_code[1];
	pkt[j++] = country_code[2];
    pkt[j++] = et->parental_rating; //-3;       //(3 less than the age)
	descriptor_length  = j-2;
	pkt[1] = descriptor_length;
    return j;
}

int set_event_information_section(type_eventtable *et, byte *pkt)
{
    int j;
    byte table_id = 0x60;
    byte section_syntax_indicator = 1;
    word section_length;
    byte version_number = 1;
    byte current_next_indicator = 0;
    byte section_number = 0;
    byte last_section_number = 0;
    byte segment_last_section_number = 0;
    byte last_table_id = 0;
    int descriptors_loop_length_pos;
    word descriptors_loop_length;
    byte running_status = 0;
    byte free_ca_mode = 0;
    dword crc_32;

    j = 0;
    pkt[j++] = table_id;
    j+=2;  //fill in section_length later
    pkt[j++] = (et->service_id & 0xff00) >> 8;
    pkt[j++] = (et->service_id & 0x00ff) >> 0;
    pkt[j++] = ((0x3) << 6)
             | ((version_number & 0x1f) << 1)
             | ((current_next_indicator & 0x1) << 0);
    pkt[j++] = section_number;
    pkt[j++] = last_section_number;

    pkt[j++] = (et->transport_stream_id & 0xff00) >> 8;
    pkt[j++] = (et->transport_stream_id & 0x00ff) >> 0;
    pkt[j++] = (et->original_network_id & 0xff00) >> 8;
    pkt[j++] = (et->original_network_id & 0x00ff) >> 0;
    pkt[j++] = segment_last_section_number;
    pkt[j++] = last_table_id;

    pkt[j++] = (et->event_id & 0xff00) >> 8;
    pkt[j++] = (et->event_id & 0x00ff) >> 0;
    pkt[j++] = (et->start_date >> 8) & 0xff;
    pkt[j++] = (et->start_date >> 0) & 0xff;
    pkt[j++] = to_bcd(et->start_hour);
    pkt[j++] = to_bcd(et->start_min);
    pkt[j++] = to_bcd(0);
    pkt[j++] = to_bcd(et->duration_hour);
    pkt[j++] = to_bcd(et->duration_min);
    pkt[j++] = to_bcd(0);
    descriptors_loop_length_pos = j;
    j+=2;   //fill in descriptors_loop_length later

    if (et->event_name) {
        j += set_short_event_descriptor(et, &pkt[j]);
    }
    if (et->num_extended && et->extended_length && et->extended_event_name) {
        j += set_extended_event_descriptor(et, &pkt[j]);
    }
    if (et->parental_rating) {
        j += set_parental_rating_descriptor(et, &pkt[j]);
    }

    //now fill in the real descriptors_loop_length
    descriptors_loop_length = j - descriptors_loop_length_pos - 2;
    pkt[descriptors_loop_length_pos] = (((et->char00 & 0x1f) & 0x7) << 5)
                                     | ((free_ca_mode & 0x1) << 4)
                                     | ((descriptors_loop_length & 0xf00) >> 8);
    pkt[descriptors_loop_length_pos+1] = (descriptors_loop_length & 0x0ff) >> 0;

    //initial crc
    pkt[j+0] = 0;
    pkt[j+1] = 0;
    pkt[j+2] = 0;
    pkt[j+3] = 0;

    //now fill in the real section_length
    section_length = j - 3 + 4;
    pkt[1] = ((section_syntax_indicator & 0x1) << 7)
           | ((0x7) << 4)
           | ((section_length & 0xf00) >> 8);
    pkt[2] = (section_length & 0x0ff) >> 0;

    crc_32 = crc32(&pkt[0], j);
    pkt[j++] = (byte)((crc_32 & 0xff000000) >> 24);
    pkt[j++] = (byte)((crc_32 & 0x00ff0000) >> 16);
    pkt[j++] = (byte)((crc_32 & 0x0000ff00) >> 8);
    pkt[j++] = (byte)((crc_32 & 0x000000ff) >> 0);

    return j;
}




//just pass it on for now
//could search for
//      [S] - subtitles -> extended + remove from short event description
//      1st sentence (if very short, no verbs)-> sub-title ?
//
int convert_short_event_descriptor(byte *from, byte *to, int *to_len)
{
/*
    int j,k;
    int descriptor_tag;
    int descriptor_length;
    int event_name_length;
    int text_length;

    j = 0;
    k = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag != SHORT_EVENT_DESCRIPTOR) {
        TAP_Print("ERROR: descriptor tag = %04x, expected 0x04x", descriptor_tag, SHORT_EVENT_DESCRIPTOR);
        return 2+descriptor_length;
    }
    j += 3;      //iso_639_language_code

    event_name_length = b70(from, j++);
    event->title = malloc(event_name_length+1);
    memcpy(event->title, &from[j], event_name_length);
    event->title[event_name_length] = '\0';
    j += event_name_length;

    text_length = b70(from, j++);
    event->desc = malloc(text_length+1);
    memcpy(event->desc, &from[j], text_length);
    event->desc[text_length] = '\0';
    j += text_length;

    *to_len = k;
    return j;
*/
    int j;
    int descriptor_tag;
    int descriptor_length;

    j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);
    j = 2+descriptor_length;
    memcpy(to, from, j);
    *to_len = j;
    return j;
}

int num_extended;

//just pass it on for now
//could search for
//      genre:  --> content_descriptor
//
int convert_extended_event_descriptor(byte *from, byte *to, int *to_len)
{
/*
    int i,j,k,l,m;
    int descriptor_tag;
    int descriptor_length;
    int descriptor_number;
    int last_descriptor_number;
    int length_of_items;
    int item_description_length;
    int item_length;
    int text_length;
    char text[0x100];
    int numin;
    int year;
    int star_rating;

    j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag != EXTENDED_EVENT_DESCRIPTOR) {
        TAP_Print("\nERROR: descriptor tag = %04x, expected 0x04x", descriptor_tag, EXTENDED_EVENT_DESCRIPTOR);
        return 2+descriptor_length;
    }
    descriptor_number          = b74(from, j);
    last_descriptor_number     = b30(from, j++);
    j += 3;       //iso_639_language_code

    l = j;
    i = 0;
    length_of_items = b70(from, j++);
    while (j < (l+length_of_items)) {
        item_description_length = b70(from, j++);
        m = j;
        j += item_description_length;
        item_length = b70(from, j++);
        //parse for
        //genre:
        //director:
        //flags:
        if (strnicmp(&from[m], "genre", 5) == 0) {
            event->genre = malloc(item_length+1);
            memcpy(event->genre, &from[j], item_length);
            event->genre[item_length] = '\0';
        }
        else
        if (strnicmp(&from[m], "director", 8) == 0) {
            event->director = malloc(item_length+1);
            memcpy(event->director, &from[j], item_length);
            event->director[item_length] = '\0';
        }
        else
        if (strnicmp(&from[m], "flags", 5) == 0) {
            //parse for
            //[R]    repeat
            //[SUB]  subtitles
            //[PREM] premiere
            //[B&W]  black&white
            //[16:9] widescreen
            //[NEW]  new_series
            //[1929] year
            //[1STARS] star_rating
            m = 0;
            while(m < item_length) {
                if (from[j+m] == '[') {
                    if (strnicmp(&from[j+m], "[R]", 3) == 0) {
                        event->repeat = 1;
                    }
                    else
                    if (strnicmp(&from[j+m], "[SUB]", 5) == 0) {
                        event->subtitles = 1;
                    }
                    else
                    if (strnicmp(&from[j+m], "[PREM]", 6) == 0) {
                        event->premiere = 1;
                    }
                    else
                    if (strnicmp(&from[j+m], "[B&W]", 5) == 0) {
                        event->black_and_white = 1;
                    }
                    else
                    if (strnicmp(&from[j+m], "[16:9]", 6) == 0) {
                        event->widescreen = 1;
                    }
                    else
                    if (strnicmp(&from[j+m], "[NEW]", 5) == 0) {
                        event->new_series = 1;
                    }
                    else
                    if (numin = sscanf(&from[j+m], "[%d]", &year)) {
                        event->year = year;
                    }
                    else
                    if (numin = sscanf(&from[j+m], "[%dSTARS]", &star_rating)) {
                        event->star_rating = star_rating;
                    }
                    else {
                    }
                }
                //skip to next []
                m++;
            }
        }
        else {
            printf("\next: ");
            for (i=0; i<item_description_length; i++) {
                printf("%c", from[m+i]);
            }
            printf(" : ");
            for (i=0; i<item_length; i++) {
                printf("%c", from[m+item_description_length+i]);
            }
        }
        j += item_length;
    }
    text_length = b70(from, j++);
    memcpy(text, &from[j], text_length);
    text[text_length] = '\0';
    j += text_length;

    *to_len = k;
    return j;
*/
    int j;
    int descriptor_tag;
    int descriptor_length;

    j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);
    j = 2+descriptor_length;
    memcpy(to, from, j);

    num_extended++;      //ugh - a global ..
    *to_len = j;
    return j;
}


//convert a component descriptor to an extended descriptor
int convert_component_descriptor(byte *from, byte *to, int *to_len)
{
    int j,k,l,m;
    int descriptor_tag;
    int descriptor_length;
    int stream_content;
    int component_type;
    int component_tag;
    int text_char_length;
    char text_char[0x100];

    byte descriptor_number = 0;
    byte last_descriptor_number = 0;
    byte iso_639_language_code[3] = {'e','n','g'};
    int length_of_items_pos;
    byte length_of_items;
    int len;
    int offset;

    j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag != COMPONENT_DESCRIPTOR) {
        TAP_Print("\nERROR: descriptor tag = %04x, expected 0x04x", descriptor_tag, COMPONENT_DESCRIPTOR);
        return 2+descriptor_length;
    }

    k = 0;
    to[k++] = EXTENDED_EVENT_DESCRIPTOR;
    k++;  //fill in descriptor_length later
    to[k++] = ((descriptor_number << 4) & 0xf0)
             | ((last_descriptor_number << 0) & 0x0f);
    to[k++] = iso_639_language_code[0];
    to[k++] = iso_639_language_code[1];
    to[k++] = iso_639_language_code[2];
    length_of_items_pos = k++;  //fill in length_of_items later


    stream_content = b30(from, j++);
    component_type = b70(from, j++);
    component_tag = b70(from, j++);
    j += 3;
    m = k;

    switch(stream_content) {
        case 1:
            len = strlen("aspect");
            to[k++] = len;
            memcpy(&to[k], "aspect", len);
            k += len;

            len = strlen(get_component_type_str(stream_content, component_type));
            to[k++] = len;
            memcpy(&to[k], get_component_type_str(stream_content, component_type), len);
            k += len;
            break;
        case 2:
            len = strlen("audio");
            to[k++] = len;
            memcpy(&to[k], "audio", len);
            k += len;

            len = strlen(get_component_type_str(stream_content, component_type));
            to[k++] = len;
            memcpy(&to[k], get_component_type_str(stream_content, component_type), len);
            k += len;
            break;
        case 3:
            len = strlen("subtitles");
            to[k++] = len;
            memcpy(&to[k], "subtitles", len);
            k += len;

            len = strlen(get_component_type_str(stream_content, component_type));
            to[k++] = len;
            memcpy(&to[k], get_component_type_str(stream_content, component_type), len);
            k += len;
            break;
        break;
            //skip
            return 2+descriptor_length;
    }
    //
    length_of_items = k-m;
    to[length_of_items_pos] = length_of_items;

    //ignore any text
    text_char_length = descriptor_length-6;
    j += text_char_length;

    //there may be some plain text
    to[k] = 0;
    k ++;

	descriptor_length  = k-2;
	to[1] = descriptor_length;

    num_extended++;      //ugh - a global ..
    *to_len = k;
    return j;
}

//convert a content descriptor to an extended descriptor
int convert_content_descriptor(byte *from, byte *to, int *to_len)
{
    int i,j,k,l,m;
   int descriptor_tag;
    int descriptor_length;
    byte descriptor_number = 0;
    byte last_descriptor_number = 0;
    byte iso_639_language_code[3] = {'e','n','g'};
    int length_of_items_pos;
    byte length_of_items;
    byte *content_str;
    int len;
    int offset;

    j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag != CONTENT_DESCRIPTOR) {
        TAP_Print("\nERROR: descriptor tag = %04x, expected 0x04x", descriptor_tag, CONTENT_DESCRIPTOR);
        return 2+descriptor_length;
    }

    k = 0;
    to[k++] = EXTENDED_EVENT_DESCRIPTOR;
    k++;  //fill in descriptor_length later
    to[k++] = ((descriptor_number << 4) & 0xf0)
             | ((last_descriptor_number << 0) & 0x0f);
    to[k++] = iso_639_language_code[0];
    to[k++] = iso_639_language_code[1];
    to[k++] = iso_639_language_code[2];
    length_of_items_pos = k++;  //fill in length_of_items later


    l = j;
    m = k;
    i = 0;
    while (j < (l+descriptor_length)) {
        //only use 1st content
        if (i == 0) {
            //ignore if genre is undefined
            content_str = get_content_type_str(b74(from, j), b30(from, j));
            if (content_str != content[0].description) {
                len = strlen("genre");
                to[k++] = len;
                memcpy(&to[k], "genre", len);
                k += len;

                len = strlen(content_str);
                to[k++] = len;
                memcpy(&to[k], content_str, len);
                k += len;
            }
        }
        j ++;
        j ++;        //user_nibble1,user_nibble2
        i++;
    }
    //
    length_of_items = k-m;
    to[length_of_items_pos] = length_of_items;

    //there may be some plain text
    to[k] = 0;
    k ++;

	descriptor_length  = k-2;
	to[1] = descriptor_length;

    num_extended++;      //ugh - a global ..
    *to_len = k;
    return j;
}

//do nothing - just pass it on
int convert_parental_rating_descriptor(byte *from, byte *to, int *to_len)
{
    int j,k;
    int descriptor_tag;
    int descriptor_length;
    byte country_code[3];
    int parental_rating;

   j = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag != PARENTAL_RATING_DESCRIPTOR) {
        TAP_Print("\nERROR: descriptor tag = %04x, expected 0x04x", descriptor_tag, PARENTAL_RATING_DESCRIPTOR);
        return 2+descriptor_length;
    }

	k = 0;
	to[k++] = descriptor_tag;
	k++;  //fill in descriptor_length later


    while (j<(descriptor_length+2)) {
	    country_code[0] = from[j++];
	    country_code[1] = from[j++];
	    country_code[2] = from[j++];
        parental_rating = b70(from, j)+3;
        j++;

	    to[k++] = country_code[0];
	    to[k++] = country_code[1];
	    to[k++] = country_code[2];
        to[k++] = parental_rating -3;       //(3 less than the age)
    }
	descriptor_length  = k-2;
	to[1] = descriptor_length;

    *to_len = k;
    return j;
}


int convert_descriptor(byte *from, byte *to, int *to_len)
{
    int j,k;
    int descriptor_tag;
    int descriptor_length;

    j = 0;
    k = 0;
    descriptor_tag = b70(from, j++);
    descriptor_length = b70(from, j++);

    if (descriptor_tag == 0)
        return 0;
    if (descriptor_length == 0)
        return 0;

    j = 0;
    switch(descriptor_tag) {
        case SHORT_EVENT_DESCRIPTOR: //0x4d
            j += convert_short_event_descriptor(from, to, &k);
            break;
        case EXTENDED_EVENT_DESCRIPTOR: //0x4e
            j += convert_extended_event_descriptor(from, to, &k);
            break;
        case COMPONENT_DESCRIPTOR: //0x50
            j += convert_component_descriptor(from, to, &k);
            break;
        case CONTENT_DESCRIPTOR: //0x54
            j += convert_content_descriptor(from, to, &k);
            break;
        case PARENTAL_RATING_DESCRIPTOR: //0x55
            j += convert_parental_rating_descriptor(from, to, &k);
            break;

        default:
            //pass on unhandled desriptors
            memcpy(&to[0], &from[0], 2+descriptor_length);
            j += 2;
            j += descriptor_length;
            k += 2;
            k += descriptor_length;
            break;
    }

    *to_len += k;
   //should check j == 2 + descriptor_length
    return 2 + descriptor_length;
}


int convert_event_information_section(byte *from, byte *to, int *to_len)
{
    char strTemp[0x40];
    int i,j,k,l;
    byte table_id;
    byte section_syntax_indicator;
    word section_length;
    byte version_number;
    byte current_next_indicator;
    byte section_number;
    byte last_section_number;
    byte segment_last_section_number;
    byte last_table_id;
    word descriptors_loop_length;
    byte running_status;
    byte free_ca_mode;

    word service_id;
    word transport_stream_id;
    word original_network_id;
    word event_id;
    byte start_date[2];
    byte start_time[3];
    byte duration[2];
    word descriptor_length;
    dword crc;
    int descriptors_loop_length_pos;

    word mjd;
    byte hour, min, sec;
    dword now;
    dword end;
    dword date;

    int k0,k1,kk;
    int discard;


#ifdef _WIN32
    SYSTEMTIME SystemTime;

    GetLocalTime(&SystemTime);
    mjd = setmjd(SystemTime.wDay, SystemTime.wMonth, SystemTime.wYear);
    now = (mjd << 16) + (SystemTime.wHour*60) + (SystemTime.wMinute);
#else
    TAP_GetTime(&mjd, &hour, &min, &sec);
    now = (mjd << 16) + (hour*60) + (min);
#endif

    j = 0;
    table_id                    = b70(from, j++);
    section_syntax_indicator    = b7(from, j);
    section_length              = b110(from, j);  j+=2;
    service_id                  = b150(from, j);  j+=2;
    version_number              = b51(from, j);
    current_next_indicator      = b0(from, j++);
    section_number              = b70(from, j++);
    last_section_number         = b70(from, j++);

    transport_stream_id         = b150(from, j);  j+=2;
    original_network_id         = b150(from, j);  j+=2;
    segment_last_section_number = b70(from, j++);
    last_table_id               = b70(from, j++);

    //fix service_id for 4tv channel5   (3241-->3242)
    if (service_id == 0x3241) {
        service_id  = 0x3242;
    }
    //fix service_id for 4tv itv3       (206d-->2066)
    if (service_id == 0x206d) {
        service_id  = 0x2066;
    }

    //fix transport_stream_id for 4tv
    if (transport_stream_id == 0) {
        transport_stream_id = service_id & 0xf000;
        transport_stream_id = (transport_stream_id == 0x1000) ? 0x1004 :
                              (transport_stream_id == 0x2000) ? 0x2005 :
                              (transport_stream_id == 0x3000) ? 0x3002 :
                              (transport_stream_id == 0x4000) ? 0x4000 :
                              (transport_stream_id == 0x5000) ? 0x5000 : 0x6000;
    }


    k = 0;
    to[k++] = table_id;
    k+=2;  //fill in section_length later
    to[k++] = (service_id & 0xff00) >> 8;
    to[k++] = (service_id & 0x00ff) >> 0;
    to[k++] = ((0x3) << 6)
             | ((version_number & 0x1f) << 1)
             | ((current_next_indicator & 0x1) << 0);
    to[k++] = section_number;
    to[k++] = last_section_number;

    to[k++] = (transport_stream_id & 0xff00) >> 8;
    to[k++] = (transport_stream_id & 0x00ff) >> 0;
    to[k++] = (original_network_id & 0xff00) >> 8;
    to[k++] = (original_network_id & 0x00ff) >> 0;
    to[k++] = segment_last_section_number;
    to[k++] = last_table_id;


    l = j;
    discard = 1;
    while (j<(3+section_length-4)) {
        event_id                            = b150(from, j);  j+=2;
        start_date[0]                       = b70(from, j++);
        start_date[1]                       = b70(from, j++);
        start_time[0]                       = b70(from, j++);
        start_time[1]                       = b70(from, j++);
        start_time[2]                       = b70(from, j++);
        duration[0]                         = b70(from, j++);
        duration[1]                         = b70(from, j++);
        duration[2]                         = b70(from, j++);
        running_status                      = b75(from, j);
        free_ca_mode                        = b4(from, j);
        descriptors_loop_length             = b110(from, j);  j+=2;

        to[k++] = (event_id & 0xff00) >> 8;
        to[k++] = (event_id & 0x00ff) >> 0;
        to[k++] = start_date[0];
        to[k++] = start_date[1];
        to[k++] = start_time[0];
        to[k++] = start_time[1];
        to[k++] = start_time[2];
        to[k++] = duration[0];
        to[k++] = duration[1];
        to[k++] = duration[2];
        descriptors_loop_length_pos = k;
        k+=2;   //fill in descriptors_loop_length later


        end = (start_date[0] << 24)+(start_date[1] << 16)
            + (from_bcd(start_time[0])*60)+(from_bcd(start_time[1]))
            + (from_bcd(duration[0])*60)+(from_bcd(duration[1]));      //what about day rollover ...
        //ignore if event is already over
        if (now > end) {
            //drop the event
            j += descriptors_loop_length;
        }
        else {
            l = j;
            k0 = k;
            num_extended = 0;            //ugh - a global ..
            while ((j < (l+descriptors_loop_length)) && (k<0xffd)) {
                k1 = k;
                j += convert_descriptor(&from[j], &to[k], &k);


                //!!make sure the section has not grown too big ...
                //if so just throw away the last few
                if ((k+4) > 0x1000) {
                    date = (start_date[0]<<8)+(start_date[1]<<0);
//                    TAP_Print("\nWARNING: section too long, service_id=%04x, %s, %02x:%02x", service_id, mjd_to_string(date), start_time[0], start_time[1]);

                    //if the last descriptor was an extended, reduce the count
                    if (to[k1] == EXTENDED_EVENT_DESCRIPTOR) {
                        num_extended --;
                    }

                    //delete the last descriptor
                    k = k1;
                }
            }


            //update the number of extended events ...
            //cant do this until all the extended descriptors have been generated
            kk = 0;
            while ((kk < num_extended) && (k0 < k)) {
                if (to[k0] == EXTENDED_EVENT_DESCRIPTOR) {
                    to[k0+2] = ((kk << 4) & 0xf0)
                             | (((num_extended-1) << 0) & 0x0f);
                    kk ++;
                }
                k0 = k0 + 2 + to[k0+1];
            }

            discard = 0;
        }

        //now fill in the real descriptors_loop_length
        descriptors_loop_length = k - descriptors_loop_length_pos - 2;
        to[descriptors_loop_length_pos] = (((running_status & 0x1f) & 0x7) << 5)
                                        | ((free_ca_mode & 0x1) << 4)
                                        | ((descriptors_loop_length & 0xf00) >> 8);
        to[descriptors_loop_length_pos+1] = (descriptors_loop_length & 0x0ff) >> 0;
    }

    j += 4;     //crc

    //initial crc
    to[k+0] = 0;
    to[k+1] = 0;
    to[k+2] = 0;
    to[k+3] = 0;

    //if all the streams were discarded, discard the whole section; by returning a duff length/crc
    if (discard) {
        *to_len = 0;
        return j;
    }

    //now fill in the real section_length
    section_length = k - 3 + 4;
    to[1] = ((section_syntax_indicator & 0x1) << 7)
           | ((0x7) << 4)
           | ((section_length & 0xf00) >> 8);
    to[2] = (section_length & 0x0ff) >> 0;

    //set the crc
    crc = crc32(&to[0], k);
    to[k+0] = (crc >> 24) & 0xff;
    to[k+1] = (crc >> 16) & 0xff;
    to[k+2] = (crc >> 8) & 0xff;
    to[k+3] = (crc >> 0) & 0xff;
    k += 4;

    *to_len = k;
    return j;
}

//----------------------------------------------------------------------------
